/*****************************************************************
 * FILE: bound.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @ALIGNALIGN_RELEASE_DATE@
 * Version:  @ALIGNALIGN_RELEASE_VERSION@
 * 
 ****************************************************************/

#include "bound.h"
#include "alignalign.h"
#include <stdlib.h>		/* NULL */

inline int
bound(aa_ushort *shape, int c, int row, int column) {
  /* 
   *Hgap, Vgap, and Dgap are upperbounds on the cost (unweighted? aka cost = number) of gaps
   * that could be open in the alignment so far and open in the
   * computation of the extension (which was done via optimistic
   * on the reverse of the input )
   */
  static int Hgap, Vgap, Dgap;
  static int i, j;

  Hgap = Vgap = Dgap = 0;  //TW - tally of how much will be subtracted from cost (below)
  for (i = 0; i < aa_K; i++) {
    for (j = 0; j < aa_L; j++) { // for each pair of seqs
      int loc_gamma;
      if (shape[i] < shape[aa_K+j]) {  //underhanging 
		/* an extension starting with deletion could count a gap twice */
		loc_gamma = GET_GAMMA_A(i,row-1);
		Hgap += aa_weights[i][aa_K+j] * loc_gamma;
		if (aa_A[i][row] == AA_EPS) {
	 		/* cannot determine this case, so add to the upperbound */
			Vgap += aa_weights[i][aa_K+j] * loc_gamma;
			Dgap += aa_weights[i][aa_K+j] * loc_gamma;
		}
      } else if (shape[i] > shape[aa_K+j]) { // overhang
		/* an extension starting with insertion could count a gap twice */
		loc_gamma = GET_GAMMA_B(j,column-1);
		Vgap += aa_weights[i][aa_K+j] * loc_gamma;
	
		if (aa_B[j][column] == AA_EPS) {
		  /* cannot determine this case, so add to the upperbound */
		  Hgap += aa_weights[i][aa_K+j] * loc_gamma;
		  Dgap += aa_weights[i][aa_K+j] * loc_gamma;
		}
      }
    }
  }

/*  printf ("to %d, %d ?\n", row, column);
  printf ("\tH:  %d - %d + %d = %d <=? %d\n" , c, Hgap, aa_optH[row][column], c - Hgap + aa_optH[row][column] , aa_cost_upper_bound);
  printf ("\tV:  %d - %d + %d = %d <=? %d\n" , c, Vgap, aa_optV[row][column], c - Vgap + aa_optV[row][column] , aa_cost_upper_bound);
  printf ("\tD:  %d - %d + %d = %d <=? %d\n" , c, Dgap, aa_optD[row][column], c - Dgap + aa_optD[row][column] , aa_cost_upper_bound);    
*/
  /*
   * if, for any of the three possible ways to begin an extension,
   * the current cost plus a lowerbound on the extension to a global
   * alignment (adjusted conservatively for gaps conted in both) is
   * no worse than the upperbound on an optimal solution, then we
   * cannot safely prune the shape.
   */   
  if ((c - Hgap + aa_optH[row][column] <= aa_cost_upper_bound ) ||
      (c - Vgap + aa_optV[row][column] <= aa_cost_upper_bound ) ||
      (c - Dgap + aa_optD[row][column] <= aa_cost_upper_bound )) {
    return AA_FALSE;
  }
  /* 
   * otherwise the shape cannot be extended to an optimal solution
   * and may be safely pruned.
   */
  else {
    return AA_TRUE;
  }
}


inline int
get_Hgap(aa_ushort *shape, int c, int row, int column) {
  static int Hgap;
  static int i, j;

  Hgap =  0.0;
  for (i = 0; i < aa_K; i++) {
    for (j = 0; j < aa_L; j++) { // for each pair of seqs
      int loc_gamma;
      if (shape[i] < shape[aa_K+j]) {  //underhanging 
		/* an extension starting with deletion could count a gap twice */
		loc_gamma = GET_GAMMA_A(i,row-1);
		Hgap += aa_weights[i][aa_K+j] * loc_gamma;
      } else if (shape[i] > shape[aa_K+j]) { // overhang
		/* and extension starting with insertion could count a gap twice */
		loc_gamma = GET_GAMMA_B(j,column-1);
		if (aa_B[j][column] == AA_EPS) {
		  /* cannot determine this case, so add to the upperbound */
		  Hgap += aa_weights[i][aa_K+j] * loc_gamma;
		}
      }
    }
  }
  return Hgap;
}


inline int
get_Vgap(aa_ushort *shape, int c, int row, int column) {
  static int Vgap;
  static int i, j;

  Vgap = 0.0;  
  for (i = 0; i < aa_K; i++) {
    for (j = 0; j < aa_L; j++) { // for each pair of seqs
      int loc_gamma;
      if (shape[i] < shape[aa_K+j]) {  //underhanging 
		/* an extension starting with deletion could count a gap twice */
		loc_gamma = GET_GAMMA_A(i,row-1);
		if (aa_A[i][row] == AA_EPS) {
	 		/* cannot determine this case, so add to the upperbound */
			Vgap += aa_weights[i][aa_K+j] * loc_gamma;
		}
      } else if (shape[i] > shape[aa_K+j]) { // overhang
		/* and extension starting with insertion could count a gap twice */
		loc_gamma = GET_GAMMA_B(j,column-1);
		Vgap += aa_weights[i][aa_K+j] * loc_gamma;
      }
    }
  }
  return Vgap;
}

inline int 
get_Dgap(aa_ushort *shape, int c, int row, int column) {
  static int Dgap;
  static int i, j;

  Dgap = 0.0;  

  for (i = 0; i < aa_K; i++) {
    for (j = 0; j < aa_L; j++) { // for each pair of seqs
      int loc_gamma;
      if (shape[i] < shape[aa_K+j]) {  //underhanging 
		/* an extension starting with deletion could count a gap twice */
		loc_gamma = GET_GAMMA_A(i,row-1);
		if (aa_A[i][row] == AA_EPS) {
	 		/* cannot determine this case, so add to the upperbound */
			Dgap += aa_weights[i][aa_K+j] * loc_gamma;
		}
      } else if (shape[i] > shape[aa_K+j]) { // overhang
		/* an extension starting with insertion could count a gap twice */
		loc_gamma = GET_GAMMA_B(j,column-1);
		if (aa_B[j][column] == AA_EPS) {
		  /* cannot determine this case, so add to the upperbound */
		  Dgap += aa_weights[i][aa_K+j] * loc_gamma;
		}
      }
    }
  }
  return Dgap;
}
