/*****************************************************************
 * FILE: driver_IO.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @ALIGNALIGN_RELEASE_DATE@
 * Version:  @ALIGNALIGN_RELEASE_VERSION@
 * 
 ****************************************************************/

#include <stdio.h>		/* fprintf */
#include <stdlib.h>		/* NULL */
#include <string.h>		/* strlen */
#include <ctype.h>		/* isspace */
#include <math.h>		/* log, ceil */

#include "alignalign.h"

/********************************************************************
 * Procedure: ReadSequences
 *   Input:  seq_file: path of file containing sequence information
 *   Output: s: array of sequences
 *           seqLength: length of longest sequence read
 *           names: array of names associated with the sequences
 *           return seq_num: number of sequences read
 *******************************************************************/
int
/*ReadSequences(char *seq_file, char ***s, int *seqLength, char ***names, 
				int **numGaps, int ***gapStarts, int ***gapLengths)
*/
ReadSequences(char *seq_file, char ***s, int *seqLength, char ***names)
{	
	
	
    FILE           *sfile;
    int             seq_num,
                    seq_len,
                    max_seq,
                    line_len,
                    i;
    char            buf[AA_MAXLINE],
                  **sq,
                  **nm;

    sfile = fopen(seq_file, "r");
    if (NULL == sfile) {
		perror(seq_file);
		exit(1);
    }
    
    seq_num = seq_len = max_seq = 0;
    while (NULL != fgets(buf, AA_MAXLINE, sfile)) {
		line_len = strlen(buf);
		if ((line_len > 1) && ('\n' == buf[line_len - 1])) {
		    buf[line_len - 1] = '\0';
		    line_len--;
		}
		if ((line_len > 1) && ('\r' == buf[line_len - 1] )) {		
 		    buf[line_len - 1] = '\0';
		    line_len--;
		}		
		if ('#' == buf[0] || '>' == buf[0]) {
		    seq_num++;
		} else if (isspace((int) buf[0])) {
		    if (seq_len > max_seq) {
				max_seq = seq_len;
		    };
		    seq_len = 0;
		} else {
		    seq_len += line_len;
		    if (seq_len > max_seq) {
				max_seq = seq_len;
		    };
		}
    }
    
    sq = malloc(seq_num * sizeof(char *));
    nm = malloc((seq_num + 1) * sizeof(char *));
    nm[seq_num] = NULL;
    for (i = 0; i < seq_num; i++) {
		sq[i] = malloc((max_seq + 1 + 1) * sizeof(char)); //TO DO one of the "+1"s was added by TW to resolve bug where next code was writing to invalid location ... don't see the bug yet.
		sq[i][0] = '\0';
		nm[i] = malloc(AA_MAXNAME * sizeof(char));
    }

    fseek(sfile, 0, SEEK_SET);
    seq_num = -1;
    seq_len = 0;
    while (NULL != fgets(buf, AA_MAXLINE, sfile)) {
		line_len = strlen(buf);
		if ((line_len > 1) && ('\n' == buf[line_len - 1] )) {
		    buf[line_len - 1] = '\0';
		    line_len--;
		}
		if ((line_len > 1) && ('\r' == buf[line_len - 1] )) {		
 		    buf[line_len - 1] = '\0';
		    line_len--;
		}				
		if ('#' == buf[0] || '>' == buf[0]) {
		    memset(nm[seq_num + 1], ' ', AA_MAXNAME - 1);
		    strncpy(nm[seq_num + 1], buf, AA_MAXNAME - 1);
		    nm[seq_num + 1][AA_MAXNAME - 1] = '\0';
		    seq_num++;
//		} else if ('\n' == buf[0]) {
		} else if (isspace((int) buf[0])) {
		    seq_len = 0;
		} else {
		    strcat(sq[seq_num], buf);
		}
    }
    seq_num++;

	*seqLength = max_seq;
    *names = nm;
    *s = sq;
    fclose(sfile);
    
   return seq_num;
}

/********************************************************************
 * Procedure: ReadCost
 *   Input:  cost_file: path of file containing cost information
 *   Output: cost: matrix of substitution costs (to be indexed by 
 *                 ascii values of given characters)
 *******************************************************************/
int
ReadCost(char *cost_file, int ***cost, char **alphabet, int **alphaconv)
{
    FILE           *dfile;
    int             alpha_size;	/* size of the aa_alphabet */
    int             line_len;	/* number of bytes read */
    int    c;		/* temporary buffer for a cost value */
    int             i,
                    j;
    char            buf[AA_MAXLINE];	/* temporary buffer for reading */
	int 			**my_cost;
	

    dfile = fopen(cost_file, "r");
    if (NULL == dfile) {
		perror(cost_file);
		exit(1);
    }

    /*
     * determine the size of the alphabet 
     */   
	*alphabet = aa_malloc(AA_MAXALPHA * sizeof(char));
	*alphaconv = aa_malloc(AA_MAXALPHA * sizeof(int));
	
    alpha_size = 0;
    fgets(buf, AA_MAXLINE, dfile);
    line_len = strlen(buf);

	for (i = 0; i< AA_MAXALPHA; i++){
		    (*alphabet)[i] = -1;
			(*alphaconv)[i] = -1;	
	}
	
    for (i = 0; '\n' != buf[i]; i++) {
		if (!isspace((int) buf[i])) {
		    (*alphabet)[alpha_size] = buf[i];
			(*alphaconv)[(int)buf[i]] = alpha_size;	
		    alpha_size++;
		}
    }

    /*
     * initialize cost matrix 
     */
	my_cost = aa_malloc(alpha_size * sizeof(int*));
	for (i=0; i<alpha_size; i++) {
		my_cost[i] = aa_malloc((alpha_size+1) * sizeof(int));
		for (j = 0; j < alpha_size; j++) {
		    my_cost[i][j] = -1.0;
		}		
	}

    /*
     * fill in the cost matrix 
     */
    fseek(dfile, 0, SEEK_SET);
    fgets(buf, AA_MAXLINE, dfile);
    for (i = 0; i < alpha_size; i++) {
		for (j = 0; j <= i; j++) {
		    fscanf(dfile, "%d", &c);
		    my_cost[i][j] = c;
		    my_cost[j][i] = c;
		}
    }

	*cost = my_cost;

    fclose(dfile);
    return alpha_size;
}


/********************************************************************
 * Procedure: ReadWeights
 *   Input:  weight_file: path of file containing weight information
 *   Output: weight: matrix of row pair weights.
 *    return T/F, if weight_file was found
 *******************************************************************/
int
ReadWeights(char *weight_file, int K, int ***weights)
{
    FILE           *wfile;
    double          w;		/* temporary buffer for a weight value */
    int          	w2;		/* temporary buffer for a weight value */    
    int             i,
                    j;
    char            buf[AA_MAXLINE];

    if (NULL == weight_file) {
    	return AA_FALSE;
    } else {
		wfile = fopen(weight_file, "r");
		if (NULL == wfile) {
		    perror(weight_file);
		    exit(1);
		}


/*		*weights = aa_malloc( K * sizeof(int*) );
		for (i = 0; i < K; i++) {
			(*weights)[i] = aa_malloc( K * sizeof(int) );
		    for (j = 0; j < K; j++) {
				(*weights)[i][j] = 1.0;
		    }
		}
*/

		/*
		 * fill in the weight matrix 
		 */
		*weights = aa_malloc( K * sizeof(int*) );		 
		fgets(buf, AA_MAXLINE, wfile);
	    for (i = 0; i < K; i++) {
			(*weights)[i] = aa_malloc( K * sizeof(int) );
			for (j = 0; j <= i; j++) {
			    fscanf(wfile, "%lf", &w);
			    w2 = (int)(.5+w);
			    (*weights)[i][j] = w2;
			    (*weights)[j][i] = w2;
			}
	    }

		fclose(wfile);
    }

/*	if (K != num_seqs) {		
			fprintf(stderr, "Number of sequences (%d) in weights file is not the same as in the sequence files (%d)\n", num_seqs, K+L);
		exit(0);
	}
*/
    return AA_TRUE;
}


void
printAlignment_fasta(int K, char **alignment, int alignlen, FILE * out, 
			int outputwidth, char **names)
{
	int i,j;
    for (i = 0; i < K ; i++) {
	    int g = strlen(names[i]);
	    if (14 <= g) {
			g = 13;
	    }
		fprintf(out, ">");
	    for (j = 1; j < g; j++) {
			fprintf(out, "%c", names[i][j]);
		}
		fprintf(out, "\n");

		int h=0;
		while (h + outputwidth < alignlen) {
			for (j = h; j < h + outputwidth; j++) {
			    fprintf(out, "%c", alignment[i][j]);
			}
		    h += outputwidth;
		    fprintf(out, "\n");
		}
		for (j = h; j < alignlen; j++) {
		    fprintf(out, "%c", alignment[i][j]);
		}
	    fprintf(out, "\n\n");


    }
}


void
printAlignment_clustal(int K, int L, char **alignment, int alignlen, FILE *out, 
				int outputwidth, char **namesA, char **namesB, unsigned char *script)		
{
    int             g,
                    h,
                    i,
                    j;
    int             current_position = 0;
    int             sequence_position[AA_MAXSEQNUM];

	int 			scriptPrinted;

    int             alignlen_digits = ceil(log10((double) alignlen));
    char            *format1;
    char            *format2;
    
    format1 = malloc(6*sizeof(char));
    format2 = malloc(8*sizeof(char));    
    
    sprintf(format1, "%%%dd  ", alignlen_digits);
    sprintf(format2, "  %%%dd  \n", alignlen_digits);



    for (i = 0; i < K + L; i++) {
		sequence_position[i] = 1;
    }

    fprintf(out, "\n");
    h = 0;
    while (h < alignlen) {
		while (h + outputwidth < alignlen) {
	    	scriptPrinted = AA_FALSE;
		    for (i = 0; i < K + L; i++) {
				if (i < K) {
				    g = strlen(namesA[i]);
				    if (14 <= g) {
						g = 13;
				    }
				    for (j = 1; j < g; j++) {
						fprintf(out, "%c", namesA[i][j]);
				    }
				    for (; j < 15; j++) {
						fprintf(out, " ");
				    }
				    fprintf(out, " ");
				} else if (i == K && scriptPrinted == AA_FALSE) {
				    for (j = 1 ; j <= 17 + alignlen_digits; j++) {
						fprintf(out, " ");
				    }
				} else {
				    g = strlen(namesB[i - K]);
				    if (14 <= g) {
						g = 13;
				    }
				    for (j = 1; j < g; j++) {
						fprintf(out, "%c", namesB[i - K][j]);
				    }
				    for (; j < 15; j++) {
						fprintf(out, " ");
				    }
				    fprintf(out, " ");
				}


				if (i == K && scriptPrinted == AA_FALSE && script != NULL) {
					//print out pipes between alignments
					for (j = h; j < h + outputwidth; j++) {
						if (script[j] == AA_SUB) {
							fprintf (out, "|");	
						}	else {
							fprintf (out, " ");	
						}
					}
					fprintf(out,"\n");
					i--;
					scriptPrinted = AA_TRUE;
				} else {
							
					fprintf(out, format1, sequence_position[i]);
					current_position = sequence_position[i];
					for (j = h; j < h + outputwidth; j++) {
					    fprintf(out, "%c", alignment[i][j]);
					    if (AA_EPS != alignment[i][j]) {
							sequence_position[i]++;
					    }
					}
					fprintf(out, format2, sequence_position[i] - 1);
				}
		    }
		    h += outputwidth;
		    fprintf(out, "\n\n");
		}
		
//	    scriptPrinted = (script == NULL) ? AA_FALSE : AA_TRUE;
	    scriptPrinted = AA_FALSE;
		for (i = 0; i < K + L; i++) {
		    if (i < K) {
				g = strlen(namesA[i]);
				if (14 <= g) {
				    g = 13;
				}
				for (j = 1; j < g; j++) {
				    fprintf(out, "%c", namesA[i][j]);
				}
				for (; j < 15; j++) {
				    fprintf(out, " ");
				}
				fprintf(out, " ");
			} else if (i == K && !scriptPrinted ) {
				    for (j = 1 ; j <= 17 + alignlen_digits; j++) {
						fprintf(out, " ");
				    }
		    } else {
				g = strlen(namesB[i - K]);
				if (14 <= g) {
				    g = 13;
				}
				for (j = 1; j < g; j++) {
				    fprintf(out, "%c", namesB[i - K][j]);
				}
				for (; j < 15; j++) {
				    fprintf(out, " ");
				}
				fprintf(out, " ");
		    }

			if (i == K && !scriptPrinted ) {
				for (j = h; j < alignlen; j++) {
					if (script[j] == AA_SUB) {
						fprintf (out, "|");	
					}	else {
						fprintf (out, " ");	
					}
				}
				fprintf(out,"\n");
				i--;
				scriptPrinted = AA_TRUE;
			} else {
	
			    fprintf(out, format1, sequence_position[i]);
			    for (j = h; j < alignlen; j++) {
					fprintf(out, "%c", alignment[i][j]);
					if (AA_EPS != alignment[i][j]) {
					    sequence_position[i]++;
					}
			    }
			    fprintf(out, format2, sequence_position[i] - 1);
			}	
		}
		h = alignlen;
		fprintf(out, "\n");
    }
    fflush(out);

	free(format1);
	free(format2);
	
    return;
}


void
printhelp(char *prog, int verbose)
{

	if (verbose == AA_TRUE) { //I set verbose to false in a call used by a script that pushes alignalign to the server ... that's all.
	    fprintf(stderr, "\nUsage: %s <-a alignment file> <-b alignment file>", prog);
	    fprintf(stderr, " [OPTION]...\n\n");
	}
	
    fprintf(stderr, "OPTIONS\n");
    fprintf(stderr, "\tall options are order independent\n\n");
    fprintf(stderr, "\t-c file, -costmatrix file\n");
    fprintf(stderr, "\t\tuse substitution costs contained in file\n\n");
    fprintf(stderr, "\t-d n\n");
    fprintf(stderr, "\t\tuse gap open cost n\n\n");
    fprintf(stderr, "\t-e n, -extension n\n");
    fprintf(stderr, "\t\tuse gap extention cost n\n\n");
    fprintf(stderr, "\t-g n1 n2, -gap n1 n2\n");
    fprintf(stderr, "\t\tuse gap open cost n1 and gap extension cost n2\n\n");
    fprintf(stderr, "\t-f filename, -f2c filename\n");
    fprintf(stderr, "\t\tsimply convert the input file filename from FASTA format to CLUSTAL format\n\n");   
    fprintf(stderr, "\t-h, -help\n");
    fprintf(stderr, "\t\tprint this document\n\n");
    fprintf(stderr, "\t-j filename, -justscore filename\n");
    fprintf(stderr, "\t\tcalculate the SP score for the input file filename\n\n");
    fprintf(stderr, "\t-o file, -outfile file\n");
    fprintf(stderr, "\t\twrite output alignment to file\n\n");
    fprintf(stderr, "\t-p\n");
    fprintf(stderr, "\t\tdo *not* print the solution alignment\n\n");
    fprintf(stderr, "\t-q, -quiet\n");
    fprintf(stderr, "\t\tdo not print summary data\n\n");
    fprintf(stderr, "\t-s, -space\n");
    fprintf(stderr, "\t\tuse the space-optimized version (at the expense of\n");
    fprintf(stderr, "\t\tcomputation time)\n\n");
    fprintf(stderr, "\t-t, -time\n");
    fprintf(stderr, "\t\tuse the time-optimized version (at the expense of\n");
    fprintf(stderr, "\t\tcomputation space)\n\n");
    fprintf(stderr, "\t-v, -verbose\n");
    fprintf(stderr, "\t\tprint summary data (default)\n");
    fprintf(stderr, "\t-w n -width n\n");
    fprintf(stderr, "\t\tformat output alignment into blocks of n columns\n\n");
//    fprintf(stderr, "\t-w file, -weight file\n");
//    fprintf(stderr, "\t\tuse pairwise sequence weights contained in file\n");
//    fprintf(stderr, "\t\t**not implemented**\n\n");
    fprintf(stderr, "\t-x\n");
    fprintf(stderr, "\t\tif printing the solution alignment, print in FASTA format\n\n");
    fprintf(stderr, "\t-y\n");
    fprintf(stderr, "\t\tif printing the solution alignment, print in CLUSTAL format (default)\n\n");

	if (verbose == AA_TRUE) {
	    fprintf(stderr, "%s-%s\t\t%s\n", AA_PROGNAME, AA_VERSION, AA_DATE);
	}

    return;
}


/*
 * Procedure: Usage (emit a simple usage message)
 */
void
Usage(char *progname)
{				/* emit usage message */
    fprintf(stderr, "Usage: %s -a <alignment A file> ", progname);
    fprintf(stderr, "-b <alignment B file> [OPTION]...\n");
    fprintf(stderr, "\t use option -help for more information\n");
    exit(1);
}
