/*****************************************************************
 * FILE: shape.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @ALIGNALIGN_RELEASE_DATE@
 * Version:  @ALIGNALIGN_RELEASE_VERSION@
 * 
 ****************************************************************/

#include <stdlib.h> /* NULL */
#include <stdio.h>  /* fprintf */
#include <string.h> /* memset */
#include "shape.h"

void
init_nodepool (void) {
  int structsize1, structsize2;

  /* size of a node is the shape plus total number of sequences
   * (minus 1, as we have 1 byte for the unsigned char pointer
   * in the Shape type declaration */
  aa_nodesize = sizeof(Shape) + ((aa_K + aa_L - 1) * sizeof(aa_ushort));

  /* aa_nodesize optimization */
  structsize1 = sizeof(S0);
  structsize2 = sizeof(S1);
  if (structsize2 - structsize1 == 2) {
    if (aa_nodesize % 2 != 0) { aa_nodesize += (2 - (aa_nodesize % 2)); }
  }
  else if (structsize2 - structsize1 == 4) {
    if (aa_nodesize % 4 != 0) { aa_nodesize += (4 - (aa_nodesize % 4)); }
  }
  else if (structsize2 - structsize1 == 8) {
    if (aa_nodesize % 8 != 0) { aa_nodesize += (8 - (aa_nodesize % 8)); }
  }
  
  /* number of nodes that will fit in MALLOCBLOCKSIZE memory */
  aa_nodemallocsize = ((int)(AA_MALLOCBLOCKSIZE / aa_nodesize));
  aa_nodemalloccount = aa_nodemallocsize;
}


/* put s on front of shapepool list */
inline void
nodefree(Shape *s) {
  NEXT(s) = aa_nodepool;
  aa_nodepool = s;
}


/* allocate a shape */
inline Shape *
nodemalloc(void) {
  static Shape *s;
  static void  **ptr;

  if (NULL == aa_nodepool) {
    if (aa_nodemalloccount == aa_nodemallocsize) {
      aa_nodeptr = aa_malloc((aa_nodesize * aa_nodemallocsize + sizeof(int *)));
      ptr = aa_nodeptr + (aa_nodesize * aa_nodemallocsize);
      *ptr = aa_nodepoolroot;
      
      aa_nodepoolroot = aa_nodeptr;
      aa_nodemalloccount = 1;
      aa_nodepoolcount++;
    }
    else {
      aa_nodemalloccount++;
      aa_nodeptr += aa_nodesize;
    }
    s = (Shape *) aa_nodeptr;
  }
  else {
    s = aa_nodepool;
    aa_nodepool = NEXT(aa_nodepool);
  }

  return s;
}


/* free the aa_nodepool */
void
nodepoolfree(int threshold) {
   void *ptr1;
   void **ptr2 = aa_malloc(sizeof(void**));
   ptr1 = aa_nodepoolroot;
   
   if (aa_nodepoolcount >= threshold) {
     while (NULL != ptr1) {
       memcpy(ptr2, (ptr1 + (aa_nodesize * aa_nodemallocsize)), sizeof(void*));
       aa_free(ptr1, (aa_nodesize * aa_nodemallocsize) + sizeof(int *));
       ptr1 = *ptr2;
     }
     aa_nodepool = NULL;
     aa_nodepoolroot = NULL;
     aa_nodeptr = NULL;
     aa_nodemalloccount = aa_nodemallocsize;
     aa_nodepoolcount = 0;
   }
   free(ptr2);
}

/**********************************************************************
 * procedure: newshape
 *   Input:
 *     aa_A, aa_B: multiple sequence alignments
 *     a, b: column indices into aa_A and aa_B respectively
 *     k, l: numbers of sequences in aa_A and aa_B respectively
 *        s: integer array representation of a shape
 *   Output: 
 *    shape: integer array representing the new shape determined by 
 *           appending colums a of aa_A and b of aa_B to s.
 *********************************************************************/
inline void
newshape (char **A, char **B, int a, int b, aa_ushort *s, 
	  aa_ushort *shape) {
  static int i, j, max;
  static int cnt[AA_MAXSEQNUM+1];/* array to store flags for block presence*/
  static int cnv[AA_MAXSEQNUM+1];/* array to compress block number range   */

  memcpy((void *) shape, (void *) s, sizeof(aa_ushort) * (aa_K+aa_L));
  memset((void *) cnt, 0, sizeof(int) * (aa_K+aa_L+1));
  memset((void *) cnv, 0, sizeof(int) * (aa_K+aa_L+1));
  max = 0;
  for (i = 0; i < aa_K+aa_L; i++) { if (max < s[i]) { max = s[i]; } }

  /* determine upper portion of new shape */
  if (A != NULL) {
    for (i = 0; i < aa_K; i++) {
      if (A[i][a] != AA_EPS)        { shape[i] = max+1; }
      else                       { shape[i] = s[i];  } } }
  //else { for (i = 0; i < aa_K; i++) { shape[i] = s[i];  } }

  /* determine lowwer portion of new shape */
  if (B != NULL) {
    for (i = 0; i < aa_L; i++) {
      if (B[i][b] != AA_EPS)        { shape[i+aa_K] = max+1;  }
      else                       { shape[i+aa_K] = s[i+aa_K]; } } }
  //else { for (i = 0; i < aa_L; i++) { shape[i+k] = s[i+aa_K]; } }

  /* compress block range */ /* set flags (cnt) for blocks present in shape */
  for (i = 0; i < aa_K+aa_L; i++) { cnt[shape[i]]++; }
  j = -1;
  for (i = 0; i < aa_K+aa_L+1; i++) {
    if (cnt[i] > 0) { j++; } /* fill in the conversion array with a unique  */
    cnv[i] = j;              /* block number for each block in shape        */
  }
  for (i = 0; i < aa_K+aa_L; i++) { shape[i] = cnv[shape[i]]; } /* set new blocks */
}


/* utility function to print a text representation of a shape */
void
printshape(aa_ushort *s) {
  int i;

  fprintf(stdout, "( "); 
  if (s) {
    for (i = 0; i < aa_K+aa_L; i++) { fprintf(stdout, "%d ", s[i]); }
  }
  else { fprintf(stdout, "NULL "); }
  fprintf(stdout, ")");
}


/* predicte returns true if input shapes are the same */
inline int 
equalshapes(aa_ushort *s1, aa_ushort *s2) {
  static int i;
  
  for (i = 0; i < aa_K+aa_L; i++) {
    if (s1[i] != s2[i]) { return AA_FALSE; }
  }
  return AA_TRUE;
}

/* count the number pairs of rows that have complimentary overhang */
inline int
compshapes(aa_ushort *s1, aa_ushort *s2) {
  static int i, j, gaps;

  gaps = 0;
  for (i = 0; i < aa_K; i++) { 
    for (j = aa_K; j < aa_K+aa_L; j++) {
      if ((s1[i] < s1[j] && s2[i] < s2[j]) || 
	  (s1[i] > s1[j] && s2[i] > s2[j])) { 
		gaps++;
      }
    }
  }
  return gaps;
}

/*
 * Procedure: reclaim
 *   Input:   n: a pointer to a shape node
 *   Output:  none
 *   Side effects: decrements the reference count to the n.  If the
 *            reference count has gone to zero, the node is freed
 *            and the process iterates on its back pointer. 
 */
inline void
reclaim(Shape *s) {
  return;
  Shape *p = s, *q = NULL;

  while (0 >= C(p)) {
    q = p;
    p = RECOVER(p);
    //if (NULL == p) { break; }
    (C(p))--;
    nodefree(q);
  }
}

