/*****************************************************************
 * FILE: Opal_IO.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @OPAL_RELEASE_DATE@
 * Version:  @OPAL_RELEASE_VERSION@
 * 
 ****************************************************************/

#include <stdio.h>		/* fprintf */
#include <stdlib.h>		/* NULL */
#include <string.h>		/* strlen */
#include <ctype.h>		/* isspace */
#include <math.h>		/* log, ceil */

#include "alignalign/alignalign.h"
#include "Opal.h"

/********************************************************************
 * Procedure: ReadSequences
 *   Input:  seq_file: path of file containing sequence information
 *   Output: s: array of sequences
 *           seqLength: length of longest sequence read
 *           names: array of names associated with the sequences
 *           return seq_num: number of sequences read
 *******************************************************************/
int
ReadSequences(char *seqFile, char ****s, int **lengths, char ***names)
{
    FILE           *sfile;
    int             seq_num,
                    seq_len,
                    line_len,
                    i;
    char            buf[AA_MAXLINE],
                  ***sq,
                  **nm;

    sfile = fopen(seqFile, "r");
    if (NULL == sfile) {
		perror(seqFile);
		exit(1);
    }

    //not very efficient ...    
    seq_num = 0;
    while (NULL != fgets(buf, AA_MAXLINE, sfile)) {
		if ('#' == buf[0] || '>' == buf[0]) {
		    seq_num++;		
	    }
    }
    
    fseek(sfile, 0, SEEK_SET);
	int * myLengths = aa_malloc(seq_num * sizeof(int));
    seq_num = seq_len = 0;
    while (NULL != fgets(buf, AA_MAXLINE, sfile)) {
		line_len = strlen(buf);
		if ((line_len > 1) && ('\n' == buf[line_len - 1] )) {		
 		    buf[line_len - 1] = '\0';
		    line_len--;
		}
		if ((line_len > 1) && ('\r' == buf[line_len - 1] )) {// why both? windows files may have both
 		    buf[line_len - 1] = '\0';
		    line_len--;
		}		
		if ('#' == buf[0] || '>' == buf[0]) {
		    seq_num++;
		    if (seq_len > 0) { //don't freak out if there are multiple newlines in a row
		    	myLengths[seq_num-1] = seq_len; // set the length of the previous seq to what was seen
		    }
		    seq_len = 0;
		} else if (! isspace((int) buf[0])) {  // skip empty lines
			//don't need to keep gap characters
			for (i=0; i<line_len; i++) {
				if (buf[i] != '-' && buf[i] != '\0'){
					seq_len++;	
				}
			}
		    if (seq_num>0) {
		    	myLengths[seq_num-1] = seq_len;
		    }
		}
    }


    
    sq = aa_malloc(seq_num * sizeof(char **));
    nm = aa_malloc((seq_num + 1) * sizeof(char *));
    nm[seq_num] = NULL;
    for (i = 0; i < seq_num; i++) {
    	//I have one entry for each sq because an sq is a group of sequences, which starts off with size 1.
		sq[i] = aa_malloc( sizeof(char *));
		sq[i][0] = aa_malloc((myLengths[i] + 1 + 1) * sizeof(char)); //TO DO one of the "+1"s was added by TW to resolve bug where next code was writing to invalid location ... don't see the bug yet.
		sq[i][0][0] = '\0';
		nm[i] = aa_malloc(AA_MAXNAME * sizeof(char));
    }



    fseek(sfile, 0, SEEK_SET);
    seq_num = -1;
    seq_len = 0;
	char tmp[2];
	tmp[1] = '\0';
    while (NULL != fgets(buf, AA_MAXLINE, sfile)) {
		line_len = strlen(buf);
		if ((line_len > 1) && ('\n' == buf[line_len - 1] )) {
		    buf[line_len - 1] = '\0';
		    line_len--;
		} 
		if ((line_len > 1) && ('\r' == buf[line_len - 1] )) {
 		    buf[line_len - 1] = '\0';
		    line_len--;
		}		
		if ('#' == buf[0] || '>' == buf[0]) {
		    memset(nm[seq_num + 1], ' ', AA_MAXNAME - 1);
		    strncpy(nm[seq_num + 1], buf, AA_MAXNAME - 1);
		    nm[seq_num + 1][AA_MAXNAME - 1] = '\0';
		    seq_num++;
		} else if (! isspace((int) buf[0])) {
			for (i=0; i<line_len; i++) {
				if (buf[i] != '-' && buf[i] != '\0'){
					tmp[0] = buf[i];
				    strcat(sq[seq_num][0], tmp);
				}
			}
		}
    }
    seq_num++;

    *lengths = myLengths;
    *names = nm;
    *s = sq;

    fclose(sfile);
    
   	return seq_num;
}

/********************************************************************
 * Procedure: ReadCost
 *   Input:  cost_file: path of file containing cost information
 *   Output: cost: matrix of substitution costs (to be indexed by 
 *                 ascii values of given characters)
 *******************************************************************/
int
ReadCost(char *cost_file, int ***cost, char **alphabet, int **alphaconv)
{
    FILE           *dfile;
    int             alpha_size;	/* size of the aa_alphabet */
    int             line_len;	/* number of bytes read */
    int    c;		/* temporary buffer for a cost value */
    int             i,
                    j;
    char            buf[AA_MAXLINE];	/* temporary buffer for reading */
	int 			**my_cost;
	

    dfile = fopen(cost_file, "r");
    if (NULL == dfile) {
		perror(cost_file);
		exit(1);
    }

    /*
     * determine the size of the alphabet 
     */   
	*alphabet = aa_malloc(AA_MAXALPHA * sizeof(char));
	*alphaconv = aa_malloc(AA_MAXALPHA * sizeof(int));
	
    alpha_size = 0;
    fgets(buf, AA_MAXLINE, dfile);
    line_len = strlen(buf);

	for (i = 0; i< AA_MAXALPHA; i++){
		    (*alphabet)[i] = -1;
			(*alphaconv)[i] = -1;	
	}
	
    for (i = 0; '\n' != buf[i]; i++) {
		if (!isspace((int) buf[i])) {
		    (*alphabet)[alpha_size] = buf[i];
			(*alphaconv)[(int)buf[i]] = alpha_size;	
		    alpha_size++;
		}
    }

    /*
     * initialize cost matrix 
     */
	my_cost = aa_malloc(alpha_size * sizeof(int*));
	for (i=0; i<alpha_size; i++) {
		my_cost[i] = aa_malloc((alpha_size+1) * sizeof(int));
		for (j = 0; j < alpha_size; j++) {
		    my_cost[i][j] = -1;
		}		
	}

    /*
     * fill in the cost matrix 
     */
    fseek(dfile, 0, SEEK_SET);
    fgets(buf, AA_MAXLINE, dfile);
    for (i = 0; i < alpha_size; i++) {
		for (j = 0; j <= i; j++) {
		    fscanf(dfile, "%d", &c);
		    my_cost[i][j] = c;
		    my_cost[j][i] = c;
		}
    }

	*cost = my_cost;


    fclose(dfile);
    return alpha_size;
}

/********************************************************************
 * Procedure: ReadWeights
 *   Input:  weight_file: path of file containing weight information
 *   Output: weight: matrix of row pair weights.
 *    return T/F, if weight_file was found
 *******************************************************************/
int
ReadWeights(char *weight_file, int K, int ***weights)
{
    FILE           *wfile;
    double 			w;		/* temporary buffer for a weight value */
    int             w2;		/* temporary buffer for a weight value */
    int             i,
                    j;
    char            buf[AA_MAXLINE];

    if (NULL == weight_file) {
    	return AA_FALSE;
    } else {
		wfile = fopen(weight_file, "r");
		if (NULL == wfile) {
		    perror(weight_file);
		    exit(1);
		}

		/*
		 * fill in the weight matrix 
		 */
		*weights = aa_malloc( K * sizeof(int*) );		 
		fgets(buf, AA_MAXLINE, wfile);
	    for (i = 0; i < K; i++) {
			(*weights)[i] = aa_malloc( K * sizeof(int) );
			for (j = 0; j <= i; j++) {
			    fscanf(wfile, "%lf", &w);
			    w2 = (int)(.5+w);
			    (*weights)[i][j] = w2;
			    (*weights)[j][i] = w2;
			}
	    }

		fclose(wfile);
    }

    return AA_TRUE;
}


void
printAlignment_fasta(int K, char **alignment, int alignlen, FILE * out, 
			int outputwidth, char **names)
{
	int i,j;
    for (i = 0; i < K ; i++) {
	    int g = strlen(names[i]);
/*	    if (14 <= g) {
*			g = 13;
*	    }
*/
	    fprintf(out, ">");
	    for (j = 1; j < g; j++) {
			fprintf(out, "%c", names[i][j]);
		}
		fprintf(out, "\n");

		int h=0;
		while (h + outputwidth < alignlen) {
			for (j = h; j < h + outputwidth; j++) {
			    fprintf(out, "%c", alignment[i][j]);
			}
		    h += outputwidth;
		    fprintf(out, "\n");
		}
		for (j = h; j < alignlen; j++) {
		    fprintf(out, "%c", alignment[i][j]);
		}
	    fprintf(out, "\n\n");


    }
}


void
printAlignment_clustal(int K, char **alignment, int alignlen, FILE * out, 
			int outputwidth, char **names)
{
    int             g,
                    h,
                    i,
                    j;
    int             current_position = 0;
    int             sequence_position[AA_MAXSEQNUM];

    int             alignlen_digits = (int)ceil(log10((double) alignlen));
    char            *format1;
    char            *format2;
    
    format1 = aa_malloc(6*sizeof(char));
    format2 = aa_malloc(8*sizeof(char));    
    
    sprintf(format1, "%%%dd  ", alignlen_digits);
    sprintf(format2, "  %%%dd  \n", alignlen_digits);

    for (i = 0; i < K; i++) {
		sequence_position[i] = 1;
    }

    fprintf(out, "\n");
    h = 0;
    while (h < alignlen) {
		while (h + outputwidth < alignlen) {
		    for (i = 0; i < K ; i++) {
			    g = strlen(names[i]);
			    if (14 <= g) {
					g = 13;
			    }
			    for (j = 1; j < g; j++) {
					fprintf(out, "%c", names[i][j]);
			    }
			    for (; j < 15; j++) {
					fprintf(out, " ");
			    }
			    fprintf(out, " ");
		
				fprintf(out, format1, sequence_position[i]);
				current_position = sequence_position[i];
				for (j = h; j < h + outputwidth; j++) {
				    fprintf(out, "%c", alignment[i][j]);
				    if (AA_EPS != alignment[i][j]) {
						sequence_position[i]++;
				    }
				}
				fprintf(out, format2, sequence_position[i] - 1);
		    }
		    h += outputwidth;
		    fprintf(out, "\n");
		}
		for (i = 0; i < K ; i++) {

			g = strlen(names[i]);
			if (14 <= g) {
			    g = 13;
			}
			for (j = 1; j < g; j++) {
			    fprintf(out, "%c", names[i][j]);
			}
			for (; j < 15; j++) {
			    fprintf(out, " ");
			}
			fprintf(out, " ");
	
		    fprintf(out, format1, sequence_position[i]);
		    for (j = h; j < alignlen; j++) {
				fprintf(out, "%c", alignment[i][j]);
				if (AA_EPS != alignment[i][j]) {
				    sequence_position[i]++;
				}
		    }
		    fprintf(out, format2, sequence_position[i] - 1);
	
		}
		h = alignlen;
		fprintf(out, "\n");
    }
    fflush(out);

    return;
}

void
printcost(int space_optimized, int gamma, int lambda, char* costFile, char* seqFile, int K, 
			char **alignment, int alignlen, char* out_file, int **weights) {

	aa_setWeights(K, weights);
	int aligncost = sp(alignment, alignlen, K);
	fprintf(stderr, "--------------------------------------------\n");
	fprintf(stderr, "%s-%s ", PROGNAME, VERSION);
	
	if (space_optimized) {
	    fprintf(stderr, "(space optimized)\n");
	} else {
	    fprintf(stderr, "(time optimized)\n");
	}
	fprintf(stderr, "Gap open cost: %d\n", gamma);
	fprintf(stderr, "Gap extension cost: %d\n", lambda);
	fprintf(stderr, "Cost matrix: %s\n", costFile);
	fprintf(stderr, "Input alignments from:\n");
	fprintf(stderr, "\t%s, %d sequences\n", seqFile, K);
	if (NULL != out_file) {
	    fprintf(stderr, "Solution alignment written to: %s\n",  out_file);
	} else {
	    fprintf(stderr, "Solution alignment written to: stdout\n");
	}
	fprintf(stderr, "Solution alignment length: %d\n", alignlen);
	fprintf(stderr, "Solution alignment cost: %d\n", aligncost);
	fprintf(stderr, "--------------------------------------------\n");
	
}
	

void
printhelp(char *prog)
{

    fprintf(stderr, "\nUsage: %s -f input_file <-a output_alignment_file>", prog);
    fprintf(stderr, " [OPTION]...\n\nOPTIONS\n");
    fprintf(stderr, "\tall options are order independent\n\n");

    fprintf(stderr, "\t-f file, --fasta_file file\n");
    fprintf(stderr, "\t\tfile containing input, in fasta format\n\n");

    fprintf(stderr, "\t-w file, --weight_file file\n");
    fprintf(stderr, "\t\tuse pairwise sequence weights contained in file\n");
    fprintf(stderr, "\t\t(default is unweighted)\n\n");

    fprintf(stderr, "\t-h, -help\n");
    fprintf(stderr, "\t\tprint this document\n\n");


    fprintf(stderr, "\tOptions for cost control:\n");   
	    fprintf(stderr, "\t\t-c file, --cost_file file\n");
	    fprintf(stderr, "\t\t\tuse substitution costs contained in file\n");
	    fprintf(stderr, "\t\t\t(default = ./BLOSUM62_base100)\n\n");	    
	
	    fprintf(stderr, "\t\t-g n, --gamma n\n");
	    fprintf(stderr, "\t\t\tuse gap-open cost n (default 60)\n\n");
	
	    fprintf(stderr, "\t\t-l n, --lambda n\n");
	    fprintf(stderr, "\t\t\tuse gap-extention cost n (default 38)\n\n");
	
	    fprintf(stderr, "\t\t-d n, --gamma_term n\n");
	    fprintf(stderr, "\t\t\tuse terminal gap-open cost n (default 15)\n\n");
	
	    fprintf(stderr, "\t\t-e n, --lambda_term n\n");
	    fprintf(stderr, "\t\t\tuse terminal gap-extention cost n (default 36)\n\n");


    fprintf(stderr, "\tOutput options: \n");   
	    fprintf(stderr, "\t\t-a file, --alignment_out_file file\n");
    	fprintf(stderr, "\t\t\twrite output alignment to file\n\n");

	    fprintf(stderr, "\t\t-o n --output_width n\n");
    	fprintf(stderr, "\t\t\tformat output alignment into blocks of n columns\n\n");
    
	    fprintf(stderr, "\t\t--output_fasta\n");
	    fprintf(stderr, "\t\t\tprint output in fasta format (default = clustal format)\n\n");
    
	    fprintf(stderr, "\t\t-m file, --merge_tree_out_file file\n");
   		fprintf(stderr, "\t\t\twrite the calculated merge tree to file\n\n");

	    fprintf(stderr, "\t\t--verbose\n");
   		fprintf(stderr, "\t\t\tprint status and summary statistics\n\n");


    fprintf(stderr, "\tOptions contolling method for distance calculation: \n");   
	    fprintf(stderr, "\t\t--dist_norm_cost n\n");
	    fprintf(stderr, "\t\t\tAlignment cost, divided by alignment length (default)\n\n");
	    
	    fprintf(stderr, "\t\t--dist_cost n\n");
	    fprintf(stderr, "\t\t\tAlignment cost, not normalized\n\n");
	
	    fprintf(stderr, "\t\t--dist_pct_id n\n");
	    fprintf(stderr, "\t\t\tPercent identity, based on input alphabet\n\n");
	
	    fprintf(stderr, "\t\t--dist_compr_pct_id n\n");
	    fprintf(stderr, "\t\t\tPercent identity, based on compressed alphabet\n\n");
	

    fprintf(stderr, "\tOptions contolling merge-tree calculation: \n");
    fprintf(stderr, "\t(see paper for details): \n");
    	
	    fprintf(stderr, "\t\t--mst n\n");
    	fprintf(stderr, "\t\t\tMinimum spanning tree (default)\n\n");

	    fprintf(stderr, "\t\t--nj n\n");
	    fprintf(stderr, "\t\t\tNeighbor joining\n\n");
	
	    fprintf(stderr, "\t\t--upgma n\n");
	    fprintf(stderr, "\t\t\tUPGMA\n\n");
	    
	    fprintf(stderr, "\t\t--mst_upgma n\n");
	    fprintf(stderr, "\t\t\tmixture of 0.9*MST and 0.1*UPGMA\n");
   	    fprintf(stderr, "\t\t\t(as used in MAFFT)\n\n");
    
    fprintf(stderr, "\tOptions contolling method for merging alignments: \n");
    
	    fprintf(stderr, "\t\t--align_optimally n\n");
	    fprintf(stderr, "\t\t\tUse exact counts algorithm (default)\n\n");

	    fprintf(stderr, "\t\t--use_pessimistic n\n");
	    fprintf(stderr, "\t\t\tUse pessimistic counts algorithm\n\n");

    fprintf(stderr, "\tPolishing options: \n");

	    fprintf(stderr, "\t\t--no_polish_internal n\n");
    	fprintf(stderr, "\t\t\tdo not perform on-the-fly polishing (default)\n\n");
	
	    fprintf(stderr, "\t\t--polish_internal n\n");
    	fprintf(stderr, "\t\t\tperform on-the-fly polishing\n\n");
      
	    fprintf(stderr, "\t\t--polish_three_cut n\n");
	    fprintf(stderr, "\t\t\tUse randomized three-cut post-polishing (default)\n\n");

	    fprintf(stderr, "\t\t-r n --random_repetitions n\n");
    	fprintf(stderr, "\t\t\tnumber of post-polish repetitions, if using three-cut\n");
	    fprintf(stderr, "\t\t\t(default = min(60,3*n) for n sequences)\n\n");

	    fprintf(stderr, "\t\t--polish_exhaustive n\n");
	    fprintf(stderr, "\t\t\tUse exhaustive two-cut post-polishing (default)\n\n");



    return;
}



/* Procedure: Usage (emit a simple usage message)*/
 
void
Usage(char *progname)
{				 
	//emit usage message 
    fprintf(stderr, "Usage: %s -f input_file <-a alignment file>", progname);
    fprintf(stderr, "\t use option --help for more information\n");
    exit(1);
}
